/*
 *     CFGTest.java  Nov 30, 2001
 *
 *     Katsuhisa Maruyama (maru@fse.cs.ritsumei.ac.jp)
 */

import jp.ac.ritsumei.cs.fse.jrt.parser.*;
import jp.ac.ritsumei.cs.fse.jrt.graphs.GraphFactory;
import jp.ac.ritsumei.cs.fse.jrt.graphs.util.*;
import jp.ac.ritsumei.cs.fse.jrt.graphs.cfg.*;
import jp.ac.ritsumei.cs.fse.jrt.parser.JavaModelFactory;
import jp.ac.ritsumei.cs.fse.jrt.parser.summary.Summary;
import jp.ac.ritsumei.cs.fse.jrt.model.*;
import jp.ac.ritsumei.cs.fse.jrt.test.FileCompare;
import java.util.*;
import java.io.*;
import junit.framework.*;

public class CFGTest extends TestCase implements GraphNodeSort, GraphEdgeSort {
    private FileWriter fileWriter;

    public CFGTest(String name) {
        super(name);
    }

    public static void main(String args[]) {
        junit.textui.TestRunner.run(suite());
    }

    protected void setUp() {
        String JDKFilePath;
        if (System.getProperty("os.name").startsWith("Windows")) {
            JDKFilePath = "c:\\jdk1.3.1\\src.jar";
        } else {
            JDKFilePath = "/usr/local/j2sdk1_3_1/src.jar";
        }
        JavaModelFactory.getInstance().setJDKFile(JDKFilePath);
    }

    protected void tearDown() {
    }

    public static Test suite() {
        TestSuite suite = new TestSuite();

        List testcases = new ArrayList();
        testcases.add(new String("./testcases"));

        List jfiles = makeTestCases(testcases);
        Iterator it = jfiles.iterator();
        while (it.hasNext()) {
            final JavaFile jfile = (JavaFile)it.next();
            suite.addTest(new CFGTest("CFG Test" + jfile.getName()) {
                protected void runTest() {
                    testOutputFile(jfile);
                }
            });
        }
        return suite;
    }

    private static List makeTestCases(List testcases) {
        List jfiles = new ArrayList();
        Iterator it = testcases.iterator();
        while (it.hasNext()) {
            String dir = (String)it.next();
            Summary.getInstance().parse(dir);

            File directory = new File(dir);
            if (directory.isDirectory()) {
                String[] names = directory.list();
                for (int i = 0; i < names.length; i++) {
                    String name = dir + File.separator + names[i];

                    File file = new File(name);
                    if (file.isFile() && name.endsWith(".java")) {
                        JavaFile jfile = JavaModelFactory.getInstance().getEachJavaFile(name);
                        jfiles.add(jfile);
                    }
                }
            }
        }
        return jfiles;
    }

    public void testOutputFile(JavaFile jfile) {
        GraphNode.clearID();
        GraphFactory.getInstance().makeCFG(jfile);
        // print(jfile);
        File file = new File(jfile.getShortFileName() + ".cfg");
        write(jfile, file);
        boolean result = FileCompare.compare(jfile.getName() + ".cfg", file.getName());
        if (result) {
            file.delete();
        }
        assertTrue(result);
    }

    private void print(JavaFile jfile) {
        JavaClass jclass = (JavaClass)jfile.getJavaClasses().get(0);
        jclass.getCCFG().print();

        Iterator it = jclass.getJavaMethods().iterator();
        while (it.hasNext()) {
            JavaMethod jmethod = (JavaMethod)it.next();
            jmethod.getCFG().print();
        }
    }

    private void write(JavaFile jfile, File file) {
        try {
            fileWriter = new FileWriter(file.getName());
            writeCFG(jfile);
            fileWriter.close();
        } catch (IOException e) { }
    }

    private void writeCFG(JavaFile jfile) {
        Iterator itc = jfile.getJavaClasses().iterator();
        while (itc.hasNext()) {
            JavaClass jclass = (JavaClass)itc.next();
            writeGraph(jclass.getCCFG());

            Iterator itm = jclass.getJavaMethods().iterator();
            while (itm.hasNext()) {
                JavaMethod jmethod = (JavaMethod)itm.next();
                writeGraph(jmethod.getCFG());
            }
        }
    }

    public void writeGraph(Graph graph) {
        writeNodes(graph);
        writeEdges(graph);
    }

    public void writeNodes(Graph graph) {
        Iterator it = graph.getNodes().iterator();
        while (it.hasNext()) {
            CFGNode node = (CFGNode)it.next();
            String info = "unknown";
            switch (node.getSort()) {
                case classEntry: info = "classEntry"; break;
                case interfaceEntry: info = "interfaceEntry"; break;
                case methodEntry: info = "methodEntry"; break;
                case constructorEntry: info = "constructorEntry"; break;

                case classExit: info = "classExit"; break;
                case interfaceExit: info = "interfaceExit"; break;
                case methodExit: info = "methodExit"; break;
                case constructorExit: info = "constructorExit"; break;

                case methodCall: info = "methodCall"; break;
                case polymorphicCall: info = "polymorphicCall"; break;
                case formalIn: info = "formalIn"; break;
                case formalOut: info = "formalOut"; break;
                case actualIn: info = "actualIn"; break;
                case actualOut: info = "actualOut"; break;

                case assignmentSt: info = "assignmentSt"; break;
                case variableDecl: info = "variableDecl"; break;
                case returnSt: info = "returnSt"; break;

                case switchLabel: info = "case"; break;
                case whileSt: info = "while"; break;
                case ifSt: info = "if"; break;
                case doSt: info = "do"; break;
                case forSt: info = "for"; break;
                case breakSt: info = "break"; break;
                case continueSt: info = "continueSt:"; break;

                case switchSt: info = "switch"; break;
                case mergeSt: info = "merge"; break;
            }
            info = node.getID() + ": " + node.getSort() + " " + info;
            write(info);
        }
    }

    public void writeEdges(Graph graph) {
        Iterator it = graph.getEdges().iterator();
        while (it.hasNext()) {
            Flow edge = (Flow)it.next();
            String info = "unknown";
            switch (edge.getSort()) {
                case trueControlFlow: info = "T"; break;
                case falseControlFlow: info = "F"; break;
                case fallThroughFlow: info = "Fall"; break;
                case parameterFlow: info = "P"; break;
            }
            if (edge.isLoopBack()) {
                info = info + "L";
            }
            info = edge.getSrcNode().getID() + " -> " + edge.getDstNode().getID()
                          + ": " + info;
            write(info);
        }
    }

    private void write(String str) {
        try {
            fileWriter.write(str + "\n");
        } catch (IOException e) { }
    }
}
